# Copyright (c) OpenMMLab. All rights reserved.
from .base_panoptic_fusion_head import \
    BasePanopticFusionHead  # noqa: F401,F403
from .heuristic_fusion_head import HeuristicFusionHead  # noqa: F401,F403
from .maskformer_fusion_head import MaskFormerFusionHead  # noqa: F401,F403
